@@ -76,11 +76,16 @@ static void bgp_connect(struct bgp_proto *p);
 static void bgp_initiate(struct bgp_proto *p);
 static void bgp_setup_listen_sk(void);
 
+
 static void
-bgp_close(struct bgp_proto *p UNUSED)
+bgp_close(struct bgp_proto *p)
 {
   ASSERT(bgp_counter);
   bgp_counter--;
+
+  if (p->cf->password)
+    sk_set_md5_auth(bgp_listen_sk, p->cf->remote_ip, NULL);
+
   if (!bgp_counter)
     {
       rfree(bgp_listen_sk);
@@ -293,11 +298,7 @@ static void
 bgp_setup_sk(struct bgp_proto *p, struct bgp_conn *conn, sock *s)
 {
   s->data = conn;
-  s->ttl = p->cf->multihop ? : 1;
-  s->rbsize = BGP_RX_BUFFER_SIZE;
-  s->tbsize = BGP_TX_BUFFER_SIZE;
   s->err_hook = bgp_sock_err;
-  s->tos = IP_PREC_INTERNET_CONTROL;
   conn->sk = s;
 }
 
@@ -325,10 +326,15 @@ bgp_connect(struct bgp_proto *p)	/* Enter Connect state and start establishing c
     s->saddr = p->local_addr;
   s->daddr = p->cf->remote_ip;
   s->dport = BGP_PORT;
+  s->ttl = p->cf->multihop ? : 1;
+  s->rbsize = BGP_RX_BUFFER_SIZE;
+  s->tbsize = BGP_TX_BUFFER_SIZE;
+  s->tos = IP_PREC_INTERNET_CONTROL;
+  s->password = p->cf->password;
+  s->tx_hook = bgp_connected;
   BGP_TRACE(D_EVENTS, "Connecting to %I from local address %I", s->daddr, s->saddr);
   bgp_setup_conn(p, conn);
   bgp_setup_sk(p, conn, s);
-  s->tx_hook = bgp_connected;
   conn->state = BS_CONNECT;
   if (sk_open(s))
     {
@@ -393,6 +399,7 @@ bgp_incoming_connection(sock *sk, int dummy UNUSED)
 		  }
 		bgp_setup_conn(p, &p->incoming_conn);
 		bgp_setup_sk(p, &p->incoming_conn, sk);
+		sk_set_ttl(sk, p->cf->multihop ? : 1);
 		bgp_send_open(&p->incoming_conn);
 		return 0;
 	      }
@@ -414,7 +421,6 @@ bgp_setup_listen_sk(void)
       s->type = SK_TCP_PASSIVE;
       s->sport = BGP_PORT;
       s->tos = IP_PREC_INTERNET_CONTROL;
-      s->ttl = 1;
       s->rbsize = BGP_RX_BUFFER_SIZE;
       s->tbsize = BGP_TX_BUFFER_SIZE;
       s->rx_hook = bgp_incoming_connection;
@@ -479,6 +485,15 @@ bgp_start_locked(struct object_lock *lock)
   p->local_id = cf->c.global->router_id;
   p->next_hop = cf->multihop ? cf->multihop_via : cf->remote_ip;
   p->neigh = neigh_find(&p->p, &p->next_hop, NEF_STICKY);
+
+  if (cf->rr_client)
+    {
+      p->rr_cluster_id = cf->rr_cluster_id ? cf->rr_cluster_id : p->local_id;
+      p->rr_client = cf->rr_client;
+    }
+
+  p->rs_client = cf->rs_client;
+
   if (!p->neigh)
     {
       log(L_ERR "%s: Invalid next hop %I", p->p.name, p->next_hop);
@@ -505,6 +520,7 @@ bgp_start(struct proto *P)
 
   bgp_counter++;
   bgp_setup_listen_sk();
+
   if (!bgp_linpool)
     bgp_linpool = lp_new(&root_pool, 4080);
 
@@ -522,6 +538,17 @@ bgp_start(struct proto *P)
   lock->hook = bgp_start_locked;
   lock->data = p;
   olock_acquire(lock);
+
+  /* We should create security association after we get a lock not to 
+   * break existing connections.
+   */
+  if (p->cf->password)
+    {
+      int rv = sk_set_md5_auth(bgp_listen_sk, p->cf->remote_ip, p->cf->password);
+      if (rv < 0)
+	return PS_STOP;
+    }
+
   return PS_START;
 }
 
@@ -611,6 +638,16 @@ bgp_check(struct bgp_config *c)
     cf_error("Local AS number must be set");
   if (!c->remote_as)
     cf_error("Neighbor must be configured");
+  if (!bgp_as4_support && c->enable_as4)
+    cf_error("AS4 support disabled globbaly");
+  if (!c->enable_as4 && (c->local_as > 0xFFFF))
+    cf_error("Local AS number out of range");
+  if (!c->enable_as4 && (c->remote_as > 0xFFFF))
+    cf_error("Neighbor AS number out of range");
+  if ((c->local_as != c->remote_as) && (c->rr_client))
+    cf_error("Only internal neighbor can be RR client");
+  if ((c->local_as == c->remote_as) && (c->rs_client))
+    cf_error("Only external neighbor can be RS client");
 }
 
 static void
