@@ -1,7 +1,7 @@
 /*
  *	BIRD -- OSPF
  *
- *	(c) 1999--2004 Ondrej Filip <feela@network.cz>
+ *	(c) 1999--2005 Ondrej Filip <feela@network.cz>
  *
  *	Can be freely distributed and used under the terms of the GNU GPL.
  */
@@ -10,6 +10,14 @@
 #define _BIRD_OSPF_H_
 
 #define MAXNETS 10
+#define OSPF_VLINK_MTU 576	/* RFC2328 - A.1 */
+#define OSPF_MAX_PKT_SIZE 65536
+			/*
+                         * RFC 2328 says, maximum packet size is 65535
+			 * This could be too much for small systems, so I
+			 * normally allocate 2*mtu - (I found one cisco
+			 * sending packets mtu+16)
+			 */
 #ifdef LOCAL_DEBUG
 #define OSPF_FORCE_DEBUG 1
 #else
@@ -192,6 +200,7 @@ struct ospf_iface
   list nbma_list;
   u8 priority;			/* A router priority for DR election */
   u8 ioprob;
+  u32 rxbuf;
 };
 
 struct ospf_md5
@@ -558,6 +567,10 @@ struct ospf_iface_patt
 #define OSPF_AUTH_SIMPLE 1
 #define OSPF_AUTH_CRYPT 2
 #define OSPF_AUTH_CRYPT_SIZE 16
+  u32 rxbuf;
+#define OSPF_RXBUF_NORMAL 0
+#define OSPF_RXBUF_LARGE 1
+#define OSPF_RXBUF_MINSIZE 256	/* Minimal allowed size */
   list *passwords;
   list nbma_list;
 };
