@@ -1,7 +1,7 @@
 /*
  *	BIRD -- OSPF
  *
- *	(c) 1999--2004 Ondrej Filip <feela@network.cz>
+ *	(c) 1999--2005 Ondrej Filip <feela@network.cz>
  *
  *	Can be freely distributed and used under the terms of the GNU GPL.
  */
@@ -31,14 +31,17 @@ ospf_pkt_fill_hdr(struct ospf_iface *ifa, void *buf, u8 h_type)
 unsigned
 ospf_pkt_maxsize(struct ospf_iface *ifa)
 {
-  return ifa->iface->mtu - SIZE_OF_IP_HEADER -
-    ((ifa->autype == OSPF_AUTH_CRYPT) ? OSPF_AUTH_CRYPT_SIZE : 0);
+  unsigned mtu = (ifa->type == OSPF_IT_VLINK) ? OSPF_VLINK_MTU : ifa->iface->mtu;
+  /* Can be mtu < 576? */
+  return ((mtu <=  ifa->iface->mtu) ? mtu : ifa->iface->mtu) -
+  SIZE_OF_IP_HEADER - ((ifa->autype == OSPF_AUTH_CRYPT) ? OSPF_AUTH_CRYPT_SIZE : 0);
+  /* For virtual links use mtu=576 */
 }
 
 void
 ospf_pkt_finalize(struct ospf_iface *ifa, struct ospf_packet *pkt)
 {
-  struct password_item *passwd = password_find (ifa->passwords);
+  struct password_item *passwd = NULL;
   void *tail;
   struct MD5Context ctxt;
   char password[OSPF_AUTH_CRYPT_SIZE];
@@ -49,6 +52,7 @@ ospf_pkt_finalize(struct ospf_iface *ifa, struct ospf_packet *pkt)
   {
     case OSPF_AUTH_SIMPLE:
       bzero(&pkt->u, sizeof(union ospf_auth));
+      passwd = password_find(ifa->passwords, 1);
       if (!passwd)
       {
         log( L_ERR "No suitable password found for authentication" );
@@ -62,6 +66,7 @@ ospf_pkt_finalize(struct ospf_iface *ifa, struct ospf_packet *pkt)
 				  sizeof(struct ospf_packet), NULL);
       break;
     case OSPF_AUTH_CRYPT:
+      passwd = password_find(ifa->passwords, 0);
       if (!passwd)
       {
         log( L_ERR "No suitable password found for authentication" );
@@ -120,7 +125,7 @@ ospf_pkt_checkauth(struct ospf_neighbor *n, struct ospf_iface *ifa, struct ospf_
       return 1;
       break;
     case OSPF_AUTH_SIMPLE:
-      pass = password_find (ifa->passwords);
+      pass = password_find(ifa->passwords, 1);
       if(!pass)
       {
         OSPF_TRACE(D_PACKETS, "OSPF_auth: no password found");
@@ -146,7 +151,7 @@ ospf_pkt_checkauth(struct ospf_neighbor *n, struct ospf_iface *ifa, struct ospf_
       }
       if (ntohs(pkt->length) + OSPF_AUTH_CRYPT_SIZE != size)
       {
-        OSPF_TRACE(D_PACKETS, "OSPF_auth: size mismatch (%d vs %s)",
+        OSPF_TRACE(D_PACKETS, "OSPF_auth: size mismatch (%d vs %d)",
 	  ntohs(pkt->length) + OSPF_AUTH_CRYPT_SIZE, size);
         return 0;
       }
@@ -162,7 +167,7 @@ ospf_pkt_checkauth(struct ospf_neighbor *n, struct ospf_iface *ifa, struct ospf_
       WALK_LIST(ptmp, *(ifa->passwords))
       {
         if (pkt->u.md5.keyid != ptmp->id) continue;
-        if ((ptmp->accfrom > now) || (ptmp->accto < now)) continue;
+        if ((ptmp->accfrom > now_real) || (ptmp->accto < now_real)) continue;
         pass = ptmp;
         break;
       }
@@ -292,9 +297,10 @@ ospf_rx_hook(sock * sk, int size)
     return 1;
   }
 
-  if ((unsigned) size > ifa->iface->mtu)
+  if (((unsigned) size > sk->rbsize) || (ntohs(ps->length) > sk->rbsize))
   {
-    log(L_ERR "%s%I - received larger packet than MTU", mesg, sk->faddr);
+    log(L_ERR "%s%I - packet is too large (%d-%d vs %d)",
+      mesg, sk->faddr, size, ntohs(ps->length), sk->rbsize);
     return 1;
   }
 
