@@ -20,6 +20,8 @@
 #define HASH_LO_STEP 2
 #define HASH_LO_MIN 8
 
+int ptp_unnumbered_stub_lsa = 0;
+
 static void *
 originate_rt_lsa_body(struct ospf_area *oa, u16 * length)
 {
@@ -28,7 +30,7 @@ originate_rt_lsa_body(struct ospf_area *oa, u16 * length)
   int j = 0, k = 0;
   u16 i = 0;
   struct ospf_lsa_rt *rt;
-  struct ospf_lsa_rt_link *ln;
+  struct ospf_lsa_rt_link *ln, *ln_after;
   struct ospf_neighbor *neigh;
 
   DBG("%s: Originating RT_lsa body for area \"%I\".\n", po->proto.name,
@@ -39,6 +41,9 @@ originate_rt_lsa_body(struct ospf_area *oa, u16 * length)
     if ((ifa->oa == oa) && (ifa->state != OSPF_IS_DOWN))
     {
       i++;
+      if ((ifa->type == OSPF_IT_PTP) && (ifa->state == OSPF_IS_PTP) &&
+      	  (ptp_unnumbered_stub_lsa || !(ifa->iface->addr->flags & IA_UNNUMBERED)))
+        i++;
     }
   }
   rt = mb_allocz(po->proto.pool, sizeof(struct ospf_lsa_rt) +
@@ -48,6 +53,7 @@ originate_rt_lsa_body(struct ospf_area *oa, u16 * length)
   if ((po->ebit) && (!oa->stub))
     rt->veb.bit.e = 1;
   ln = (struct ospf_lsa_rt_link *) (rt + 1);
+  ln_after = ln + i;
 
   WALK_LIST(ifa, po->iface_list)
   {
@@ -61,6 +67,9 @@ originate_rt_lsa_body(struct ospf_area *oa, u16 * length)
     if ((ifa->oa != oa) || (ifa->state == OSPF_IS_DOWN))
       continue;
 
+    if (ln == ln_after)
+      die("LSA space overflow");
+
     if (ifa->state == OSPF_IS_LOOP)
     {
       ln->type = 3;
@@ -81,7 +90,7 @@ originate_rt_lsa_body(struct ospf_area *oa, u16 * length)
 	  ln->id = neigh->rid;
 	  ln->metric = ifa->cost;
 	  ln->notos = 0;
-	  if (ifa->iface->flags && IA_UNNUMBERED)
+	  if (ifa->iface->addr->flags & IA_UNNUMBERED)
 	  {
 	    ln->data = ifa->iface->index;
 	  }
@@ -92,19 +101,30 @@ originate_rt_lsa_body(struct ospf_area *oa, u16 * length)
 	}
 	else
 	{
-	  if (ifa->state == OSPF_IS_PTP)
+          ln--;
+	  i--;		/* No link added */
+        }
+
+	if ((ifa->state == OSPF_IS_PTP) && 
+	    (ptp_unnumbered_stub_lsa || !(ifa->iface->addr->flags & IA_UNNUMBERED)))
+	{
+	  ln++;
+	  if (ln == ln_after)
+            die("LSA space overflow");
+
+	  ln->type = LSART_STUB;
+	  ln->metric = ifa->cost;
+	  ln->notos = 0;
+	  if (ifa->iface->addr->flags & IA_UNNUMBERED)
 	  {
-	    ln->type = LSART_STUB;
-	    ln->id = ln->id = ipa_to_u32(ifa->iface->addr->opposite);
-	    ln->metric = ifa->cost;
-	    ln->notos = 0;
+	    ln->id = ipa_to_u32(ifa->iface->addr->opposite);
 	    ln->data = 0xffffffff;
 	  }
 	  else
-          {
-            ln--;
-	    i--;		/* No link added */
-          }
+	  {
+	    ln->data = ipa_to_u32(ipa_mkmask(ifa->iface->addr->pxlen));
+	    ln->id = ipa_to_u32(ifa->iface->addr->prefix) & ln->data;
+	  }
 	}
 	break;
       case OSPF_IT_BCAST:
@@ -861,6 +881,44 @@ ospf_hash_delete(struct top_graph *f, struct top_hash_entry *e)
   bug("ospf_hash_delete() called for invalid node");
 }
 
+static void
+ospf_dump_lsa(struct top_hash_entry *he, struct proto *p)
+{
+  struct ospf_lsa_rt *rt = NULL;
+  struct ospf_lsa_rt_link *rr = NULL;
+  struct ospf_lsa_net *ln = NULL;
+  u32 *rts = NULL;
+  u32 i, max;
+
+  OSPF_TRACE(D_EVENTS, "- %1x %-1I %-1I %4u 0x%08x 0x%04x %-1I",
+	he->lsa.type, he->lsa.id, he->lsa.rt, he->lsa.age,
+	he->lsa.sn, he->lsa.checksum, he->oa ? he->oa->areaid : 0 );
+
+  switch (he->lsa.type)
+    {
+    case LSA_T_RT:
+      rt = he->lsa_body;
+      rr = (struct ospf_lsa_rt_link *) (rt + 1);
+
+      for (i = 0; i < rt->links; i++)
+        OSPF_TRACE(D_EVENTS, "  - %1x %-1I %-1I %5u", rr[i].type, rr[i].id, rr[i].data, rr[i].metric);
+      break;
+
+    case LSA_T_NET:
+      ln = he->lsa_body;
+      rts = (u32 *) (ln + 1);
+      max = (he->lsa.length - sizeof(struct ospf_lsa_header) -                                                                                                               
+		sizeof(struct ospf_lsa_net)) / sizeof(u32);
+
+      for (i = 0; i < max; i++)
+        OSPF_TRACE(D_EVENTS, "  - %-1I", rts[i]);
+      break;
+
+    default:
+      break;
+    }
+}
+
 void
 ospf_top_dump(struct top_graph *f, struct proto *p)
 {
@@ -869,14 +927,9 @@ ospf_top_dump(struct top_graph *f, struct proto *p)
 
   for (i = 0; i < f->hash_size; i++)
   {
-    struct top_hash_entry *e = f->hash_table[i];
-    while (e)
-    {
-      OSPF_TRACE(D_EVENTS, "- %1x %-1I %-1I %4u 0x%08x 0x%04x %-1I",
-		 e->lsa.type, e->lsa.id, e->lsa.rt, e->lsa.age,
-		 e->lsa.sn, e->lsa.checksum, e->oa ? e->oa->areaid : 0 );
-      e = e->next;
-    }
+    struct top_hash_entry *e;
+    for (e = f->hash_table[i]; e != NULL; e = e->next)
+      ospf_dump_lsa(e, p);
   }
 }
 
